-- ToME - Tales of Maj'Eyal
-- Copyright (C) 2009, 2010, 2011, 2012 Nicolas Casalini
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Nicolas Casalini "DarkGod"
-- darkgod@te4.org

require "engine.class"
local DamageType = require "engine.DamageType"
local Map = require "engine.Map"
local Chat = require "engine.Chat"
local Target = require "engine.Target"
local Talents = require "engine.interface.ActorTalents"

local _M = loadPrevious(...)

local base_checkHit = _M.checkHit

--Choose what to print in verbose log: LOS to player(normal) or all for cheat mode
local verbosable = function(who)
	return game.log and config.settings.tome.game_mechanics.verbose_game_log and (config.settings.cheat or who:hasLOS(game.player.x,game.player.y))
end

--New, simpler checkHit that relies on rescaleCombatStats() being used elsewhere
function _M:checkHitLin(atk, def, min, max, factor, p)
	if atk < 0 then atk = 0 end
	if def < 0 then def = 0 end
	local min = min or 0
	local max = max or 100
	if game.player:hasQuest("tutorial-combat-stats") then
		min = 0
		max = 100
	end --ensures predictable combat for the tutorial
	print("checkHit", atk, def)
	local hit = math.ceil(50 + 2.5 * (atk - def))
	hit = util.bound(hit, min, max)
	print("=> chance to hit", hit)
	return rng.percent(hit), hit
end

-- I5 updated to smooth out hit probabilites and to be more forgiving of level differences
function _M:checkHitAlt(atk, def, min, max, factor, p)
	local min = min or 0
	local max = max or 100
		
	if config.settings.cheat and type(factor) == "number" then -- tracking deprecated check-hit function calls
		local st = debug.getinfo(2,"nlS")
		print(("checkHit(smooth) received factor : %f from %s line %d in %s"):format(factor,st.name or "unknown function",st.currentline or 0,st.source or "unknown src"))
		if config.settings.cheat then
			game.log("#YELLOW#checkHit(smooth) received factor : %f from %s line %d in %s",factor,st.name or "unknown function",st.currentline or 0,st.source or "unknown src")
		end
	end
	
	if config.settings.tome.game_mechanics.smooth_hit_factor <= 0 then return _M:checkHitLin(atk,def, min, max, factor, p) end
	
	local factor = type(factor) == "table" and factor.tune or config.settings.tome.game_mechanics.smooth_hit_factor or 10
	
	local m = (factor or 0) > 0 and factor or 10
	if game.player:hasQuest("tutorial-combat-stats") then
		min = 0
		max = 100
	end --ensures predictable combat for the tutorial
	local hit = 50+50*(atk-def)/(math.abs(atk-def)+factor) -- tunable parameter in denominator 10> ~ 16.7% hit at -20 atk-def
	hit = util.bound(hit, min, max)  -- Limit checks aren't needed with the new formula.  Left in for compatibility
	
--	if config.settings.tome.game_mechanics.verbose_game_log and game.log and self:hasLOS(game.player.x,game.player.y) then
	if verbosable(self) then
		local _,oldhit = _M:checkHitLin(atk, def, min, max, factor, p)
		game.log("#LIGHT_RED#Attack: %d vs Defense: %d ==> chance to hit(smooth) is %0.1f%% (vs linear %0.1f%%)",atk,def,hit,oldhit)
	end
	print("checkHit(smooth)",atk,def,"=> chance to hit", hit)
	return rng.percent(hit), hit
end

--- Computes a logarithmic chance to hit, opposing chance to hit to chance to miss
-- This will be used for melee attacks, physical and spell resistance
function _M:checkHitAltOld(atk, def, min, max, factor)

	if config.settings.cheat and type(factor) == "number" then 
		local st = debug.getinfo(2,"nlS")
		print(("checkHitOld received factor : %f from %s line %d in %s"):format(factor,st.name or "unknown function",st.currentline or 0,st.source or "unknown src"))
		if config.settings.cheat then
			game.log("#YELLOW#checkHitOld received factor : %f from %s line %d in %s",factor,st.name or "unknown function",st.currentline or 0,st.source or "unknown src")
		end
	end
	
	if config.settings.tome.game_mechanics.smooth_hit_factor <= 0 then return _M:checkHitOldExp(atk,def, min, max, factor, p) end
	
	local factor = type(factor) == "table" and factor.tune or config.settings.tome.game_mechanics.smooth_hit_factor*2 or 20
		
--	if config.settings.tome.game_mechanics.verbose_game_log and game.log and self:hasLOS(game.player.x,game.player.y) then
	if verbosable(self) then
		local _,oldhit = _M:checkHitOldExp(atk, def, min, max, factor)
		config.settings.tome.game_mechanics.verbose_game_log=false
		local _,hit = _M:checkHit(atk, def, min, max, {tune = factor})
		config.settings.tome.game_mechanics.verbose_game_log=true
		game.log("#LIGHT_RED#Power: %d vs Save: %d ==> chance to hit(old,smooth) is %0.1f%% (vs old %0.1f%%)",atk,def,hit,oldhit)
		return _,hit
	end
	return _M:checkHit(atk, def, min, max, {tune = factor}) --I5 Redirect to normal checkHit function
end

--- Computes a logarithmic chance to hit, opposing chance to hit to chance to miss
-- This will be used for melee attacks, physical and spell resistance
function _M:checkHitOldExp(atk, def, min, max, factor)
	if atk < 0 then atk = 0 end
	if def < 0 then def = 0 end
	print("checkHit", atk, def)
	if atk == 0 then atk = 1 end
	local hit = nil
	factor = factor or 5

	local one = 1 / (1 + math.exp(-(atk - def) / 7))
	local two = 0
	if atk + def ~= 0 then two = atk / (atk + def) end
	hit = 50 * (one + two)

	hit = util.bound(hit, min or 5, max or 95)
	print("=> chance to hit", hit)
	return rng.percent(hit), hit
end

--- Try to totally evade an attack
function _M:checkEvasionAlt(target)
	if not target:attr("evasion") or self == target then return end

	local evasion = target:attr("evasion")
	print("checkEvasion", evasion, target.level, self.level)
	if not config.settings.tome.game_mechanics.limited_evasion then
		evasion = evasion * (target.level / self.level) -- base level adjustment
	end
	print("=> evasion chance", evasion)
	return rng.percent(evasion)
end

return _M